using System;
using System.Collections.Generic;
using System.Text;

using LibUsbDotNet;
using LibUsbDotNet.Info;
using LibUsbDotNet.Main;

using System.IO;
using System.Threading;

namespace HIDUSBTest
{

    /// <summary>
    /// Class to read internal and external temperatures for "HIDTEMPerNTC" USB devices.
    /// 
    /// Device manufacturer/sale: http://www.pcsensor.com/Thermometer/1-outer-NTC-sensor-thermometer/prod_27.html
    /// Also available from http://usb.brando.com/usb-indoor-outdoor-thermometer-ii_p01244c035d15.html
    /// Code based on the Relavak's source: http://relavak.wordpress.com/2009/10/17/temper-temperature-sensor-linux-driver/
    /// and analyzing/reverse-engineering the original application.
    /// 
    /// Released by Robin B. Jensen (Deep Thought), http://www.drunkardswalk.dk, on 20100102.
    /// 
    /// The code is tested on Windows XP, using LibUsbDotNet 2.1.0.
    /// Note that the libUsb-win32 filter service (included with LibUsbDotNet) apparently must be installed 
    /// for the device detection and hence the code to work properly.
    /// The code is somewhat cleaned up, but there is still quite a lot that can be added or modfied -
    /// such as error checking, comments, timer/thread-based event reporting, sanity checks, validation,... 
    /// - but I release it now for ya'all to use, then you can just expand or change what you want.
    /// 
    /// I have tested the external sensor at room temperature (21 degrees C) and in hot water (app. 80 deg. C),
    ///  and the reported values are equal to what the original application reports. 
    /// 
    /// Usage restrictions: Feel free to use the code for anything you want. 
    /// I'd appreciate a link to my website ( http://www.drunkardswalk.dk ) , though.
    /// 
    /// Disclaimer:
    /// I take no responsibility for any loss of data, system problems or any other issues which 
    /// may in any way be caused directly or indirectly by the use of this code.
    /// 
    /// </summary>
    public class HIDTEMPerNTCInterface
    {
        private UsbDevice MyUsbDevice;
        UsbDeviceFinder MyUsbFinder = new UsbDeviceFinder(0x1130, 0x660C);

        private decimal[] Res = null;//new decimal[0xd3];

        private const Int32 REQUIREDDELAY = 32;//Seems like the device needs some time to process a command before sending the next

        const Int32 DELAYFORDISASSEMBLYPURPOSE = 0;//Used while figuring out the protocol

        byte NTCGain = 1;

        const decimal DIVISOR = 52 + (1 / 30.75m);//The byte array returned from GetNTCVolt is apparently divided by this value

        const double I1 = 4.1666666666666669E-06;

        private double _offset = 0;
        public double Offset
        {
            get {
                return _offset;
            }
            set
            {
                _offset = value;
            }
        }

        public const double ABSOLUTEZERO = -273.15;

        public bool Connect()
        {
            bool found = false;

            UsbRegistry myUsbRegistry = UsbGlobals.AllDevices.Find(MyUsbFinder);

            if (ReferenceEquals(myUsbRegistry, null))
            {
                // The device is not connected or cannot be accessed by libusbdotnet
                Log("Device not connected!");
                //ShowLastUsbError();
                return found;
            }

            // Find and open the usb device.
            MyUsbDevice = UsbGlobals.OpenUsbDevice(MyUsbFinder);

            // If the device is open and ready
            if (MyUsbDevice == null)
            {
                Log("Device not found or locked :(");
                return found;
            }
            found = true;

            //Read and parse reference values
            //this.getTxtString("NTC_10K_1.txt");
            ReadReferenceValues();

            Init();

            return found;
        }

        private void ReadReferenceValues()
        {
            string datafileName = "NTC_10K_1.txt";

            string datafilePath = System.Windows.Forms.Application.StartupPath + @"\" + datafileName;

            string[] lines = File.ReadAllLines(datafilePath);

            List<decimal> tmpList = new List<decimal>();

            for (int f = 0; f < lines.Length; f++)
            {
                if (!lines[f].StartsWith("*"))
                { 
                    char[] splitAround=new char[]{';'};
                    string[] lineparts = lines[f].Split(splitAround);

                    decimal dec=Decimal.Parse(lineparts[1], System.Globalization.CultureInfo.InvariantCulture);
                    tmpList.Add(dec);
                }
            }

            Res = tmpList.ToArray();

        }

        /// <summary>
        /// Emulates EntryAPI.EMyInitConfig and FirstConfigGain
        /// </summary>
        private void Init()
        {
            List<byte> byteList = new List<byte>();

            byteList.Clear();
            byteList.AddRange(new byte[] { 10, 11, 12, 13, 0, 0, 2, 0 });
            SendControlMessage(byteList.ToArray());

            byteList.Clear();
            byteList.AddRange(new byte[] { 0x43, 0, 0, 0, 0, 0, 0, 0 });
            SendControlMessage(byteList.ToArray());

            for (int f = 0; f < 7; f++)
            {
                byteList.Clear();
                byteList.AddRange(new byte[] { 0, 0, 0, 0, 0, 0, 0, 0 });
                SendControlMessage(byteList.ToArray());
            }


            SetNTCGain(NTCGain);

        }

        /// <summary>
        /// Sends control message to device
        /// </summary>
        /// <param name="byteArray"></param>
        private void SendControlMessage(byte[] byteArray)
        {
            List<byte> byteList = new List<byte>();
            byteList.AddRange(byteArray);
            byteList.AddRange(new byte[] { 0, 0, 0, 0, 0, 0, 0, 0 });
            byteList.AddRange(new byte[] { 0, 0, 0, 0, 0, 0, 0, 0 });
            byteList.AddRange(new byte[] { 0, 0, 0, 0, 0, 0, 0, 0 });

            byte[] bytesToSend = byteList.ToArray();

            UsbSetupPacket packet = new UsbSetupPacket(UsbRequestType.TypeClass | UsbRequestType.RecipInterface, DeviceRequestType.SetConfiguration);// UsbRequestType.
            packet.Value = 0x200;
            packet.Index = 1;
            packet.Length = 32;

            Int32 numBytesTransferred = 0;

            bool sendResult = MyUsbDevice.ControlTransfer(ref packet, bytesToSend, bytesToSend.Length, out numBytesTransferred);

            if (numBytesTransferred != 32)
            {
                string result = string.Format("Packet sending result: {0} Bytes sent: {1}", sendResult, numBytesTransferred);
                Log(result);
            }

            //Console.WriteLine("Packet content:");
            //for (int f = 0; f < bytesToSend.Length; f++)
            //{
            //    Console.Write("{0:X2} ", bytesToSend[f]);

            //}
            //Console.WriteLine();
            //Console.WriteLine("---------------------------------");

            Thread.Sleep(REQUIREDDELAY);
        }

        /// <summary>
        /// Reads data control message from device
        /// </summary>
        /// <returns></returns>
        private byte[] GetDataControlMessage()
        {

            byte[] byteArray = new byte[256];

            UsbSetupPacket packet = new UsbSetupPacket(UsbRequestType.TypeClass | UsbRequestType.RecipInterface | UsbRequestType.EndpointIn, DeviceRequestType.ClearFeature);// UsbRequestType.
            packet.Value = 0x300;
            packet.Index = 1;
            packet.Length = 256;

            Int32 numBytesTransferred = 0;

            bool sendResult = MyUsbDevice.ControlTransfer(ref packet, byteArray, byteArray.Length, out numBytesTransferred);

            if (numBytesTransferred != byteArray.Length)
            {
                string result = string.Format("Packet sending result: {0} Bytes received: {1}", sendResult, numBytesTransferred);
                Log(result);
            }
            return byteArray;
        }

        /// <summary>
        /// Dummy method
        /// </summary>
        /// <param name="str"></param>
        private void Log(string str)
        {
            Console.WriteLine("Log message:" + str);
        }

        public double GetInternalTemperature()
        {
            double internalTemperature = 0;

            List<byte> byteList = new List<byte>();

            // Get temps
            byteList.Clear();
            byteList.AddRange(new byte[] { 10, 11, 12, 13, 0, 0, 2, 0 });
            SendControlMessage(byteList.ToArray());

            byteList.Clear();
            byteList.AddRange(new byte[] { 0x54, 0, 0, 0, 0, 0, 0, 0 });
            SendControlMessage(byteList.ToArray());

            for (int f = 0; f < 7; f++)
            {
                byteList.Clear();
                byteList.AddRange(new byte[] { 0, 0, 0, 0, 0, 0, 0, 0 });
                SendControlMessage(byteList.ToArray());
            }

            byteList.Clear();
            byteList.AddRange(new byte[] { 10, 11, 12, 13, 0, 0, 1, 0 });
            SendControlMessage(byteList.ToArray());

            byte[] resultArray = GetDataControlMessage();

            //for (int f = 0; f < resultArray.Length; f++)
            //{
            //    Console.Write("{0:G} ", resultArray[f]);

            //}
            //Console.WriteLine();

            internalTemperature = resultArray[0] + resultArray[1] / 256.0;

            return internalTemperature;
        }

        /// <summary>
        /// Emulates SetNTCGain in DLL
        /// </summary>
        /// <param name="gain">New gain : 1,2,4,8,16,32,64,128</param>
        private void SetNTCGain(Int32 gain)
        {
            Log("Changing gain to " + gain);

            Int32 gainForPacket = Log2(gain);

            gainForPacket += 0x61;

            //string result = string.Format("{0} transmorphed to {1}", gain, gainForPacket);
            //Console.WriteLine(result);

            List<byte> byteList = new List<byte>();


            byteList.Clear();
            byteList.AddRange(new byte[] { 10, 11, 12, 13, 0, 0, 2, 0 });
            SendControlMessage(byteList.ToArray());

            byteList.Clear();
            byteList.AddRange(new byte[] { (byte)gainForPacket, 0, 0, 0, 0, 0, 0, 0 });
            SendControlMessage(byteList.ToArray());

            for (int f = 0; f < 7; f++)
            {
                byteList.Clear();
                byteList.AddRange(new byte[] { 0, 0, 0, 0, 0, 0, 0, 0 });
                SendControlMessage(byteList.ToArray());
            }

            //And again:
            byteList.Clear();
            byteList.AddRange(new byte[] { 10, 11, 12, 13, 0, 0, 2, 0 });
            SendControlMessage(byteList.ToArray());

            byteList.Clear();
            byteList.AddRange(new byte[] { (byte)gainForPacket, 0, 0, 0, 0, 0, 0, 0 });
            SendControlMessage(byteList.ToArray());

            for (int f = 0; f < 7; f++)
            {
                byteList.Clear();
                byteList.AddRange(new byte[] { 0, 0, 0, 0, 0, 0, 0, 0 });
                SendControlMessage(byteList.ToArray());

            }
        }

        /// <summary>
        /// Emulates the DLL's GetNTCVolt()
        /// </summary>
        /// <returns></returns>
        private double GetNTCVolt()
        {
            double returnValue = 0;

            List<byte> byteList = new List<byte>();


            byteList.Clear();
            byteList.AddRange(new byte[] { 10, 11, 12, 13, 0, 0, 2, 0 });
            SendControlMessage(byteList.ToArray());

            byteList.Clear();
            byteList.AddRange(new byte[] { 0x41, 0, 0, 0, 0, 0, 0, 0 });
            SendControlMessage(byteList.ToArray());

            for (int f = 0; f < 7; f++)
            {
                byteList.Clear();
                byteList.AddRange(new byte[] { 0, 0, 0, 0, 0, 0, 0, 0 });
                SendControlMessage(byteList.ToArray());

            }

            byteList.Clear();
            byteList.AddRange(new byte[] { 10, 11, 12, 13, 0, 0, 1, 0 });
            SendControlMessage(byteList.ToArray());

            byte[] resultArray = GetDataControlMessage();

            //for (int f = 0; f < resultArray.Length; f++)
            //{
            //    Console.Write("{0:G} ", resultArray[f]);

            //}
            //Console.WriteLine();

            if (resultArray[0] == resultArray[1] && (resultArray[0] == 0 || resultArray[0] == 255))
            {
                returnValue = -2.46;
                Log("Forcing value to -2.46");
            }
            else
            {
                returnValue = resultArray[0] - 128;
                double tmp = (double)resultArray[1] / 256.0;
                //returnValue += resultArray[1] / 256;
                returnValue += tmp;
                //returnValue /= 52.0;
                returnValue = (double)((decimal)returnValue / DIVISOR);
            }
            Log("GetNTCVolt returns " + returnValue);
            return returnValue;

        }

        /// <summary>
        /// Helper method - finds Log2 of an integer
        /// </summary>
        /// <param name="d"></param>
        /// <returns></returns>
        private Int32 Log2(Int32 d)
        {
            return (Int32)(Math.Log(d) / Math.Log(2));
        }

        public bool checkGain(double volt)
        {
            Log("Checkgain with value " + volt);
            if (volt > ((2.46 / ((double)this.NTCGain)) * 0.9))
            {
                if (this.NTCGain <= 1)
                {
                    return false;
                }
                //this.NTCGain = (byte)Math.Round((double)(((double)this.NTCGain) / 2.0));
                Int32 curGain = NTCGain/2;
                NTCGain = (byte)curGain;
                Log("EntryAPI.SetNTCGain :" + this.NTCGain);
                //EntryAPI.SetNTCGain(this.NTCGain);
                SetNTCGain(this.NTCGain);
                Log("Done");
                System.Threading.Thread.Sleep(DELAYFORDISASSEMBLYPURPOSE);

            }
            else if (volt < ((2.46 / ((double)this.NTCGain)) * 0.4))
            {
                if (this.NTCGain >= 0x80)
                {
                    return false;
                }
                this.NTCGain = (byte)(this.NTCGain * 2);
                Log("EntryAPI.SetNTCGain :" + this.NTCGain);
                //EntryAPI.SetNTCGain(this.NTCGain);
                SetNTCGain(this.NTCGain);
                Log("Done");
                //Thread.Sleep(128);
                System.Threading.Thread.Sleep(DELAYFORDISASSEMBLYPURPOSE);
                //EntryAPI.EMyGetNTCVolt();
            }
            else
            {
                return false;
            }
            return true;
        }

        private double Volt7705Calibration(double volit, byte gain)
        {
            if (volit <= 0.0010888)
            {
                return -0.000334633;
            }
            if (volit <= 0.0012803)
            {
                return this.voltDengBiFengdian(volit, 0.0010888, 0.0012803, 0.000754167, 0.0009208333);
            }
            if (volit <= 0.0017002)
            {
                return this.voltDengBiFengdian(volit, 0.0012803, 0.0017002, 0.0009208333, 0.0012541667);
            }
            if (volit <= 0.002666)
            {
                return this.voltDengBiFengdian(volit, 0.0017002, 0.002666, 0.0012541667, 0.0021125);
            }
            if (volit <= 0.00522)
            {
                return this.voltDengBiFengdian(volit, 0.002666, 0.00522, 0.0021125, 0.0041666667);
            }
            if (volit <= 0.0149)
            {
                return this.voltDengBiFengdian(volit, 0.00522, 0.0149, 0.0041666667, 0.012625);
            }
            if (volit <= 0.04683)
            {
                return this.voltDengBiFengdian(volit, 0.0149, 0.04683, 0.012625, 0.0413333333);
            }
            if (volit <= 0.21342)
            {
                return this.voltDengBiFengdian(volit, 0.04683, 0.21342, 0.0413333333, 0.2115416667);
            }
            if (volit <= 0.36914)
            {
                return this.voltDengBiFengdian(volit, 0.21342, 0.36914, 0.2115416667, 0.346166667);
            }
            if (volit <= 0.44121)
            {
                return this.voltDengBiFengdian(volit, 0.36914, 0.44121, 0.346166667, 0.4215416667);
            }
            if (volit <= 0.65351)
            {
                return this.voltDengBiFengdian(volit, 0.44121, 0.65351, 0.4215416667, 0.6208333333);
            }
            if (volit <= 0.92445)
            {
                return this.voltDengBiFengdian(volit, 0.65351, 0.92445, 0.6208333333, 0.91625);
            }
            if (volit <= 1.08022)
            {
                return this.voltDengBiFengdian(volit, 0.92445, 1.08022, 0.91625, 1.1375);
            }
            if (volit <= 1.8745)
            {
                return this.voltDengBiFengdian(volit, 1.08022, 1.8745, 1.1375, 1.91375);
            }
            if (volit <= 1.9943)
            {
                return this.voltDengBiFengdian(volit, 1.8745, 1.9943, 1.91375, 2.07125);
            }
            if (volit <= 2.4589)
            {
                return this.voltDengBiFengdian(volit, 1.9943, 2.4589, 2.07125, 2.72125);
            }
            return 0.26235000000000008;
        }

        private double voltDengBiFengdian(double cureentVolt, double qiShiDianDianYa, double zhongDianDianYa, double qiShiDianLiLunZhi, double zhongDianLiLunZhi)
        {
            double liLunCeShiDianDianYaGenJuQiShiDianSuoDeZhi = (qiShiDianLiLunZhi * cureentVolt) / qiShiDianDianYa;
            double LiLunCeShiDianDianYaGenJuZhongDianSuoDeZhi = (zhongDianLiLunZhi * cureentVolt) / zhongDianDianYa;
            double liangDianZhiJianDeChaZhi = LiLunCeShiDianDianYaGenJuZhongDianSuoDeZhi - liLunCeShiDianDianYaGenJuQiShiDianSuoDeZhi;
            double gaiDianSuoZhanDeBiLi = (cureentVolt - qiShiDianDianYa) / (zhongDianDianYa - qiShiDianDianYa);
            double retunValue = liLunCeShiDianDianYaGenJuQiShiDianSuoDeZhi + (liangDianZhiJianDeChaZhi * gaiDianSuoZhanDeBiLi);
            return (retunValue - cureentVolt);
        }

        public double GetExternalTemperature()
        {
            double externalTemperature = ABSOLUTEZERO;

            double Temp_NTC = 0;//Current resistance of the NTC 

            Log("EntryAPI.EMyGetNTCVolt");
            //double getVolt = EntryAPI.EMyGetNTCVolt();
            double getVolt = GetNTCVolt();
            Log("Result = " + getVolt);

            if (getVolt == -2.46)
            {
                return ABSOLUTEZERO;//Error reading, return invalid temp;
            }

            getVolt /= (double)this.NTCGain;

            bool isChangedGain = checkGain(getVolt);

            while (isChangedGain)
            {
                Log("EntryAPI.EMyGetNTCVolt");
                //getVolt = EntryAPI.EMyGetNTCVolt();
                getVolt = GetNTCVolt();
                Log("Result = " + getVolt);
                if (getVolt == -2.46)
                {
                    return ABSOLUTEZERO;//Error reading, return invalid temp;
                }
                System.Threading.Thread.Sleep(DELAYFORDISASSEMBLYPURPOSE);
                getVolt = getVolt / ((double)this.NTCGain);
                isChangedGain = checkGain(getVolt);
            }

            double checkVolt = this.Volt7705Calibration(getVolt, 1);
            
            getVolt += checkVolt;
            
            Temp_NTC = (getVolt / I1) / 1000.0;

            if ((decimal)Temp_NTC > Res[0] || (decimal)Temp_NTC < Res[Res.Length-1])
            {
                return ABSOLUTEZERO;//FUBAR data, abort
            }

            for (int f = Res.Length-1; f >0 ; f--)
            {
                if ((decimal)Temp_NTC <= (this.Res[f]))
                {
                    //Linear interpolation
                    decimal fixPointDiff = Res[f] - Res[f + 1];
                    decimal deltaR = Res[f] - (decimal)Temp_NTC ;
                    decimal exTemp = deltaR / fixPointDiff;
                    exTemp += f - 50;
                    Console.WriteLine("exTemp:"+exTemp);
                    externalTemperature = (double)exTemp;
                    break;
                }
            
            }

            return externalTemperature;
        }

        private void ShowLastUsbError()
        {
            Console.WriteLine("Error Number: {0}", UsbGlobals.LastErrorNumber);
            Console.WriteLine(UsbGlobals.LastErrorString);
        }
    }
}
